"
 COPYRIGHT (c) 1992 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Link subclass:#Process
         instanceVariableNames:'id prio state startBlock name 
                                restartable interruptActions 
                                exitAction singleStepping
                                emergencySignalHandler'
         classVariableNames:''
         poolDictionaries:''
         category:'Kernel-Processes'
!

Process comment:'
COPYRIGHT (c) 1992 by Claus Gittinger
             All Rights Reserved

$Header: /cvs/stx/stx/libbasic/Process.st,v 1.14 1994-08-22 12:11:52 claus Exp $
'!

!Process class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1992 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /cvs/stx/stx/libbasic/Process.st,v 1.14 1994-08-22 12:11:52 claus Exp $
"
!

documentation
"
    Instances of Process represent lightweight smalltalk processes 
    (i.e. threads). These all run in a shared smalltalk/X address space,
    and can thus access and communicate via any objects.
    Do not confuse these with (heavy-weight) unix processes, which are
    created differently, and do NOT run in the same address space.
    Also notice, that heavy-weight process creation takes much longer.
    (see OperatingSystemclass>>fork).

    Processes are typically created by sending #fork or #forkAt: to a block.
    Scheduling is done by Processor, which is the sole instance of 
    ProcessorScheduler.

    Notice: in Smalltalk/X, processes are gone, when an image is restarted;
    this means, that you have to take care of process re-creation yourself.
    Usually, this is done by depending on ObjectMemory, recreating the
    process(s) when the #returnFromSnapshot-change notifiaction arrives.

    A later version will allow specification of automatic restart, but
    thats not yet implemented.

    Instance variables:

        id                     <SmallInteger>   a unique process-id
        prio                   <SmallInteger>   the processes priority
        state                  <Symbol>         the processes state
                                                (for process-monitor)
        startBlock             <Block>          the startup-block
        name                   <String-or-nil>  the processes name (if any)
                                                (for process-monitor)
        restartable            <Boolean>        is restartable (not yet implemented)
        interruptActions       <Collection>     interrupt actions (not yet tested)
        exitAction             <Block>          action to perform on termination
        emergencySignalHandler <Block>          can be used for per-process
                                                emergency signal handling
"
! !

!Process methodsFor:'accessing'!

state
    "return a symbol describing the processes state"

    ^ state
!

state:aSymbol
    "set the state - only to be used from scheduler"

    state := aSymbol
!

startBlock:aBlock
    "set the receivers startup block"

    startBlock := aBlock
!

emergencySignalHandler:aOneArgBlock
    "set the emergencySignalHandler block.
     See Signal>>documentation for more info."

    emergencySignalHandler := aOneArgBlock
!

emergencySignalHandler
    "return the emergencySignalHandler block.
     See Signal>>documentation for more info."

    ^ emergencySignalHandler
!

priority
    "return the receivers priority"

    ^ prio
!

priority:aNumber
    "set my priority"

    Processor changePriority:aNumber for:self.
!

changePriority:aNumber
    "same as priority:, but returns the old priority.
     (cannot do this in priority: for ST-80 compatibility)"

    |oldPrio|

    oldPrio := prio.
    Processor changePriority:aNumber for:self.
    ^ oldPrio
!

isSingleStepping
    ^ singleStepping
!

singleStep:aBoolean
    singleStepping := aBoolean
!

setPriority:aNumber
    "set priority without telling processor - no public use"

    prio := aNumber
!

id
    "return the processes id"

    ^ id
!

name
    "return the processes name"

    ^ name
!

name:aString
    "set the processes name"

    name := aString
!

nameOrId
    "return a string to identify the process - either name or id"

    name notNil ifTrue:[^ name].
    ^ id printString
!

exitAction
    "return the processes exit action"

    ^ exitAction
!

exitAction:aBlock
    "set the processes exit action"

    exitAction := aBlock
!

suspendedContext
    "return the processes suspended context"

%{  /* NOCONTEXT */
    extern OBJ __threadContext();
    OBJ i;

    if (_isSmallInteger(i = _INST(id))) {
        RETURN (__threadContext(_intVal(i)));
    }
%}
.
    ^ nil
!

usedStackSize
    "Return the processes current stack size.
     This method is for monitoring purposes only."

%{  /* NOCONTEXT */
    extern int __threadUsedStackSize();
    OBJ i;

    if (_isSmallInteger(i = _INST(id))) {
        RETURN( _MKSMALLINT(__threadUsedStackSize(_intVal(i))) );
    }
%}
.
    ^ nil
!

totalStackSize
    "return the processes maximum used stack size.
     This method is for monitoring purposes only."

%{  /* NOCONTEXT */
    extern int __threadTotalStackSize();
    OBJ i;

    if (_isSmallInteger(i = _INST(id))) {
        RETURN( _MKSMALLINT(__threadTotalStackSize(_intVal(i))) );
    }
%}
.
    ^ nil
!

maximumStackSize
    "returns the processes stack limit - i.e. the process will be 
     interrupted with a recursionSignal-raise, if it ever
     needs more stack (in bytes) than this number"

%{  /* NOCONTEXT */
    extern int __threadMaxStackSize();
    OBJ i;

    if (_isSmallInteger(i = _INST(id))) {
        RETURN( _MKSMALLINT(__threadMaxStackSize(_intVal(i))) );
    }
%}
.
    ^ nil
!
 
setMaximumStackSize:limit
    "sets the processes stack limit - i.e. the process will be
     interrupted with a recursionSignal-raise, if it ever
     needs more stack (in bytes) than this number.
     Returns the old value."

%{  /* NOCONTEXT */
    extern int __threadSetMaxStackSize();
    OBJ i;

    if (_isSmallInteger(i = _INST(id)) 
     && _isSmallInteger(limit) ) {
        RETURN ( _MKSMALLINT(__threadSetMaxStackSize(_intVal(i), _intVal(limit))) );
    }
%}
.
    ^ nil
! !

!Process methodsFor:'private scheduler access'!

setId:aNumber
    "set the processes id"

    id := aNumber
!

start
    "start the process (sent from VM)"

    startBlock notNil ifTrue:[
        startBlock value.
    ]
! !

!Process methodsFor:'suspend / resume'!

suspend
    "suspend the receiver process - will continue to run when a resume is sent"

    Processor suspend:self
!

resume
    "resume the receiver process"

    Processor resume:self
!

resumeForSingleSend
    "resume the receiver process, but only let it execute a single send."

    Processor resumeForSingleSend:self
!

terminate
    "terminate the receiver process"

    Processor terminate:self
! !

!Process methodsFor:'special'!

interruptWith:aBlock
    "interrupt the receiver and make it evaluate aBlock.
     If the receiver is currently suspended, the block will be remembered
     to be evaluated once the receiver wakes up."

    |wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    interruptActions isNil ifTrue:[
        interruptActions := OrderedCollection with:aBlock.
    ] ifFalse:[
        interruptActions addLast:aBlock.
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
    Processor scheduleForInterrupt:self.
!

interrupt
    "
     evaluate my interrupt-actions
     the process will go back to where it got interrupted
     after doing this.
    "
    |action|

    [interruptActions notNil and:[interruptActions notEmpty]] whileTrue:[
        action := interruptActions removeFirst.
        action value
    ].
    interruptActions := nil
! !
    
!Process methodsFor:'printing & storing'!

printOn:aStream
    aStream nextPutAll:state article;
            space;
            nextPutAll:state;
            nextPutAll:' Process (';
            nextPutAll:self nameOrId;
            nextPutAll:')'
! !
